/* #Specification: aclmod_file / concept
	This aclfs module checks if a file exists. It returns -1 if the
	file does not exist, which invalidates the ACL context.

	We are using this to enable some access control by creating a semaphore
	file somewhere in the hierarchy. Generally, these files will be created
	in /var/run for example, but they could be created by any users.

	For example, the owner of a hierarchy may revoke all access by creating
	or erasing this file. For sure, the superuser must setup the ..acl
	this way.

	#
	module aclmod_file file=/var/run/acl_web
	#
*/
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <time.h>
#include <syslog.h>
#include "aclfsd.h"

class CHKFILE{
	bool debug;
	char *file;
	int last_ret;
	time_t last_tested;		// Avoid checking the file too often.
							// We check it every 5 seconds
	bool valid;
	/*~PROTOBEG~ CHKFILE */
public:
	CHKFILE (int argc, const char *args[]);
	bool isvalid (void);
	int test (void);
	~CHKFILE (void);
	/*~PROTOEND~ CHKFILE */
};

PUBLIC CHKFILE::CHKFILE (int argc, const char *args[])
{
	debug = false;
	file = NULL;
	last_ret = -1;
	last_tested = 0;
	valid = true;
	/* #Specification: aclmod_file / arguments
		Arguments passed to module will be named so we can have optional
		arguments. This supports enhancement as well. So the syntax is

		#
		module aclmod_file debug=yes file=path
		#

		file is a path name. The module only check for the existance
		of this file to grant access.
	*/
	for (int i=0; i<argc; i++){
		const char *arg = args[i];
		if (strncmp(arg,"debug=",6)==0){
			debug = strcmp(arg+6,"yes")==0;
		}else if (strncmp(arg,"file=",5)==0){
			file = strdup(arg+5);
		}else{
			syslog (LOG_ERR,"aclmod_file: Invalid argument %s",arg);
			valid = false;
		}
	}
}

PUBLIC CHKFILE::~CHKFILE ()
{
	free (file);
}

PUBLIC bool CHKFILE::isvalid ()
{
	return valid && file != NULL;
}

PUBLIC int CHKFILE::test()
{
	time_t now = time(NULL);
	if (now - last_tested > 5){
		last_tested = now;
		struct stat st;
		if (stat(file,&st)!=-1){
			last_ret = 0;
		}else{
			last_ret = -1;
		}
	}
	if (debug) fprintf (stderr,"aclmod_file %s: ret=%d\n",file,last_ret);
	return last_ret;
}


extern "C" void *initctx (int argc, const char *args[])
{
	void *ret = NULL;
	CHKFILE *cr = new CHKFILE(argc,args);
	if (cr->isvalid()){
		ret = cr;
	}else{
		delete cr;
	}
	return ret;
}

extern "C" void endctx (void *stuff)
{
	if (stuff != NULL){
		CHKFILE *cr = (CHKFILE*)stuff;
		delete cr;
	}
}

extern "C" int testctx (
	void *stuff,
	ACL_CREDINFO &,
	const char *,		// dir
	const char *,		// file
	OPER_CONTEXT&,		// oper
	int &rights)
{
	int ret = -1;
	if (stuff == NULL){
		// Disable access since initctx could not parse
		rights = 0;
	}else{
		CHKFILE *cr = (CHKFILE*)stuff;
		ret = cr->test();
	}
	return ret;
}
