/* #Specification: aclmod_sched / concept
	This aclfs module implements some access control based on some
	scheduling. It accepts a "cron" like syntax, where one can
	specify when access is granted.

	Ultimatly, modules may be very complex, full of features. Instead
	we are shouting for very small module, so anyone can create one
	or customise to achieve a goal.

	For example, this module could very well have the following features

	-an argument would tell if the access must be granted or not granted
	 when the schedule is matching.

	-We could specify which privilege must be reset when the schedule is
	 met or not.

	-logging control.

	Future will tell.
*/
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <syslog.h>
#include "aclfsd.h"

class CRON_RANGE{
public:
	CRON_RANGE *next;
	int from;
	int to;
	/*~PROTOBEG~ CRON_RANGE */
public:
	CRON_RANGE (int _from, int _to, CRON_RANGE *_next);
	/*~PROTOEND~ CRON_RANGE */
};

PUBLIC CRON_RANGE::CRON_RANGE(int _from, int _to, CRON_RANGE *_next)
{
	next = _next;
	from = _from;
	to = _to;
}

class CRON{
	time_t last_tested;	// To avoid testing and testing
						// we only retest every 10 seconds
	int last_ret;	// Last result
	bool debug;
	bool valid;
public:
	CRON_RANGE *minutes;
	CRON_RANGE *hours;
	CRON_RANGE *mdays;
	CRON_RANGE *wdays;
	CRON_RANGE *months;
	/*~PROTOBEG~ CRON */
public:
	CRON (int argc, const char *args[]);
	void dump (void);
	bool isvalid (void);
	int test (void);
	~CRON (void);
	/*~PROTOEND~ CRON */
};

static CRON_RANGE *cron_parse(const char *s)
{
	CRON_RANGE *ret = NULL;
	if (s!=NULL && s[0] != '\0' && strcmp(s,"*")!=0){
		while (*s != '\0'){
			int from = atoi(s);
			while (isdigit(*s)) s++;
			if (*s == '-'){
				s++;
				int to = atoi(s);
				while (isdigit(*s)) s++;
				ret = new CRON_RANGE(from,to,ret);
			}else{
				ret = new CRON_RANGE(from,from,ret);
			}
			if (*s == '|') s++;
		}
	}
	return ret;
}

PUBLIC CRON::CRON(
	int argc,
	const char *args[])
{
	/* #Specification: aclmod_sched / arguments
		Arguments passed to module will be named so we can have optional
		arguments. This supports enhancement as well. So the syntax is

		#
		module aclmod_sched debug=yes units=ranges
		#

		units is one of minutes, hours, mdays (days of the month
		from 1 to 31, months (from 1 to 12) and wdays (days of the week
		from 1 to 7).

		Failing to supply one unit is equivalent to supplying * in
		the crontab syntax.

		Ranges is a set of enumaration and ranges like this: 1,2,4-6

		Interval like 1-10/3 are not supported
	*/
	debug = false;
	last_tested = 0;
	last_ret = -1;
	minutes = hours = mdays = months = wdays = NULL;
	valid = true;
	for (int i=0; i<argc; i++){
		const char *arg = args[i];
		if (strncmp(arg,"debug=",6)==0){
			debug = strcmp(arg+6,"yes")==0;
		}else if (strncmp(arg,"minutes=",8)==0){
			minutes = cron_parse (arg+8);
		}else if (strncmp(arg,"hours=",6)==0){
			hours = cron_parse (arg+6);
		}else if (strncmp(arg,"mdays=",6)==0){
			mdays = cron_parse (arg+6);
		}else if (strncmp(arg,"months=",7)==0){
			months = cron_parse (arg+7);
		}else if (strncmp(arg,"wdays=",6)==0){
			wdays = cron_parse (arg+6);
		}else{
			syslog (LOG_ERR,"aclmod_sched: Invalid argument %s",arg);
			valid = false;
		}
	}
}

/*
	Return true if the arguments were valid
*/
PUBLIC bool CRON::isvalid()
{
	return valid;
}

static void cron_delrange(CRON_RANGE *r)
{
	while (r != NULL){
		CRON_RANGE *next = r->next;
		delete r;
		r = next;
	}
}

PUBLIC CRON::~CRON()
{
	cron_delrange (minutes);
	cron_delrange (hours);
	cron_delrange (mdays);
	cron_delrange (wdays);
	cron_delrange (months);
}

/*
	Check if a value is member of a range.
	An empty range (r==NULL) match every value (* in crontab)
*/
static bool cron_chkrange (CRON_RANGE *r, int val)
{
	bool ret = true;
	if (r != NULL){
		ret = false;
		while (r != NULL){
			if (r->from <= val && r->to >= val){
				ret = true;
				break;
			}
			r = r->next;
		}
	}
	return ret;
}

static void cron_dump (const char *title, CRON_RANGE *cr)
{
	fprintf (stderr,"\t%10s:",title);
	if (cr == NULL){
		fprintf (stderr," *\n");
	}else{
		while (cr != NULL){
			if (cr->from != cr->to){
				fprintf (stderr," %d-%d",cr->from,cr->to);
			}else{
				fprintf (stderr," %d",cr->from);
			}
			cr = cr->next;
		}
		fprintf (stderr,"\n");
	}
}

PUBLIC void CRON::dump()
{
	fprintf (stderr,"Dumping aclmod_sched\n");
	cron_dump ("minutes",minutes);
	cron_dump ("hours",hours);
	cron_dump ("mdays",mdays);
	cron_dump ("months",months);
	cron_dump ("wdays",wdays);
	fprintf (stderr,"------------\n");
}

/*
	Check if we are in a valid period
*/
PUBLIC int CRON::test()
{
	int ret = -1;
	time_t now = time(NULL);
	if (now - last_tested > 10){
		last_tested = now;
		struct tm *t = localtime (&now);
		t->tm_mon ++;
		t->tm_wday ++;
		if (cron_chkrange(minutes,t->tm_min)
			&& cron_chkrange(hours,t->tm_hour)
			&& cron_chkrange(months,t->tm_mon)){
			if (mdays != NULL && wdays != NULL){
				if (cron_chkrange(wdays,t->tm_wday)
					&& cron_chkrange(mdays,t->tm_mday)){
					ret = 0;
				}
			}else if (mdays != NULL){
				if(cron_chkrange(mdays,t->tm_mday)){
					ret = 0;
				}
			}else if (wdays != NULL){
				if (cron_chkrange(wdays,t->tm_wday)){
					ret = 0;
				}
			}else{
				ret = 0;
			}
		}
		if (debug){
			fprintf (stderr,"aclmod_sched test: min=%d hour=%d mday=%d month=%d wday=%d\n"
				,t->tm_min,t->tm_hour,t->tm_mday,t->tm_mon,t->tm_wday);
			dump();
			fprintf (stderr,"ret = %d\n",ret);
		}
		last_ret = ret;
	}else{
		ret = last_ret;
	}
	return ret;
}


extern "C" void *initctx (int argc, const char *args[])
{
	void *ret = NULL;
	#if 0
		fprintf (stderr,"argc %d:",argc);
		for (int i=0; i<argc; i++) fprintf (stderr," :%s:",args[i]);
		fprintf (stderr,"\n");
	#endif
	CRON *cr = new CRON(argc,args);
	if (cr->isvalid()){
		ret = cr;
	}else{
		delete cr;
	}
	return ret;
}

extern "C" void endctx (void *stuff)
{
	if (stuff != NULL){
		// fprintf (stderr,"aclmod_sched: freeing stuff %p\n",stuff);
		CRON *cr = (CRON*)stuff;
		delete cr;
		// fprintf (stderr,"aclmod_sched: freeing stuff done\n");
	}
}

extern "C" int testctx (
	void *stuff,
	ACL_CREDINFO &,
	const char *,		// dir
	const char *,		// file
	OPER_CONTEXT&,		// oper
	int &rights)
{
	int ret = -1;
	if (stuff == NULL){
		// Disable access since initctx could not parse
		rights = 0;
	}else{
		CRON *cr = (CRON*)stuff;
		ret = cr->test();
	}
	return ret;
}
