#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include <dirent.h>
#include <syslog.h>
#include <sys/time.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include "../litefs/litefs.h"
#include "../litefs/litefs.p"
#include "../litefs/litefsutil.h"
#include "../litefs/bufinput.h"

static bool debug=false;

void n_debug (const char *ctl, ...)
{
	if (debug){
		va_list list;
		va_start (list,ctl);
		vfprintf (stderr,ctl,list);
		va_end (list);
	}
}


static void usage()
{
	fprintf (stderr,"litefsd-browse version 1.0\n");
}

class NAMEOBJ{
public:
	char *name;
	/*~PROTOBEG~ NAMEOBJ */
public:
	NAMEOBJ (const char *_name);
	virtual ~NAMEOBJ (void);
	/*~PROTOEND~ NAMEOBJ */
};

PUBLIC NAMEOBJ::NAMEOBJ(const char *_name)
{
	name = strdup(_name);
}
PUBLIC VIRTUAL NAMEOBJ::~NAMEOBJ()
{
	free (name);
}

class FILEDESC: public NAMEOBJ{
public:
	int uid;
	int gid;
	int size;
	/*~PROTOBEG~ FILEDESC */
public:
	FILEDESC (const char *_name,
		 int _uid,
		 int _gid,
		 int _size);
	/*~PROTOEND~ FILEDESC */
};

PUBLIC FILEDESC::FILEDESC(const char *_name, int _uid, int _gid, int _size)
	: NAMEOBJ(_name)
{
	uid = _uid;
	gid = _gid;
	size = _size;
}

class NAMEOBJS{
public:
	NAMEOBJ **tb;
	int nb;
	int maxtb;
	/*~PROTOBEG~ NAMEOBJS */
public:
	NAMEOBJS (void);
	void add (NAMEOBJ *obj);
	void clean (void);
	NAMEOBJ *lookup (const char *name);
	~NAMEOBJS (void);
	/*~PROTOEND~ NAMEOBJS */
};

PUBLIC NAMEOBJS::NAMEOBJS()
{
	tb = NULL;
	nb = 0;
	maxtb = 0;
}

PUBLIC NAMEOBJS::~NAMEOBJS()
{
	clean();
	free (tb);
}
PUBLIC void NAMEOBJS::clean()
{
	for (int i=0; i<nb; i++) delete tb[i];
	nb = 0;
}

PUBLIC void NAMEOBJS::add (NAMEOBJ*obj)
{
	if (nb == maxtb){
		maxtb += 100;
		tb = (NAMEOBJ**)realloc(tb,maxtb*sizeof(NAMEOBJ*));
	}
	tb[nb++] = obj;
}

PUBLIC NAMEOBJ *NAMEOBJS::lookup (const char *name)
{
	NAMEOBJ *ret = NULL;
	for (int i=0; i<nb; i++){
		NAMEOBJ *n = tb[i];
		if (strcmp(n->name,name)==0){
			ret = n;
			break;
		}
	}
	return ret;
}


class DIRDESC: public NAMEOBJ{
public:
	NAMEOBJS subdirs;
	NAMEOBJS files;
	time_t last_loaded;
	int ttl;		// Time to live
	/*~PROTOBEG~ DIRDESC */
public:
	DIRDESC (const char *_name);
	void load (const char *path);
	void loadif (const char *path);
	DIRDESC *lookupadddir (const char *dname);
	DIRDESC *lookupdir (const char *name);
	FILEDESC *lookupfile (const char *name);
	~DIRDESC (void);
	/*~PROTOEND~ DIRDESC */
};

PUBLIC DIRDESC::DIRDESC(const char *_name)
	: NAMEOBJ (_name)
{
	last_loaded = 0;
	ttl = 2;
}

PUBLIC DIRDESC::~DIRDESC()
{
}

PUBLIC DIRDESC *DIRDESC::lookupdir (const char *name)
{
	return (DIRDESC*)subdirs.lookup(name);
}

PUBLIC FILEDESC *DIRDESC::lookupfile (const char *name)
{
	return (FILEDESC*)files.lookup(name);
}

PUBLIC DIRDESC *DIRDESC::lookupadddir (const char *dname)
{
	DIRDESC *ret = lookupdir (dname);
	if (ret == NULL){
		ret = new DIRDESC (dname);
		ret->ttl = ttl;
		subdirs.add (ret);
	}
	return ret;
}


#if 0
static void strip_end (char *buf)
{
	int last = strlen(buf)-1;
	while (last >= 0 && isspace(buf[last])) buf[last--] = '\0';
}
#endif

static const char *browsecmd = "virtualfs-browsenet";

PUBLIC void DIRDESC::load(const char *path)
{
	subdirs.clean();
	files.clean();
	char cmd[PATH_MAX];
	n_debug ("loading directory :%s: (%s)\n",path,name);
	snprintf (cmd,sizeof(cmd)-1,"%s --list %s",browsecmd,path);
	FILE *fin = popen (cmd,"r");
	if (fin != NULL){
		char buf[1000];
		while (fgets(buf,sizeof(buf)-1,fin)!=NULL){
			char type[100],sname[PATH_MAX];
			int uid,gid,size;
			if (buf[0] == 't'){
				char *pt = buf+1;
				while (isspace (*pt)) pt++;
				ttl = atoi(pt);
			}else if (sscanf (buf,"%s %s %d %d %d",type,sname,&uid,&gid,&size)==5){
				if (type[0] == 'f'){
					n_debug ("file :%s: in dir :%s:\n",sname,name);
					FILEDESC *d = new FILEDESC (sname,uid,gid,size);
					files.add (d);
				}else if (type[0] == 'd'){
					n_debug ("Adding sub-directory :%s:\n",sname);
					char *pt = strchr(sname,'/');
					if (pt != NULL){
						*pt++ = '\0';
						DIRDESC *d = lookupadddir (sname);
						d->last_loaded = time(NULL);
						d->lookupadddir (pt);
						n_debug ("Add :%s: :%s:\n",sname,pt);
					}else{
						lookupadddir (sname);
					}
				}
			}
		}
		pclose (fin);
		last_loaded = time(NULL);
	}
}

/*
	Load the directory content if needed
*/
PUBLIC void DIRDESC::loadif(const char *path)
{
	time_t now = time(NULL);
	if (now - last_loaded > ttl){
		load(path);
	}
}


static DIRDESC hosts("");

static void browse_sendstat (
	bool is_file,
	LITEFSD_CTX &ctx,
	int uid,
	int gid,
	int size)
{
	struct stat st;
	// Fake a directory
	memset (&st,0,sizeof(st));
	st.st_uid = uid;
	st.st_gid = gid;
	st.st_size = size;
	st.st_mode = is_file ? 0100600 : 0040700;
	st.st_nlink = 1;
	st.st_atime = st.st_mtime = st.st_ctime = time(NULL);
	litefsd_sendstat (ctx,st);
}

struct BROWSE_DIR {
public:
	DIRDESC *dir;
	int pos;
	BROWSE_DIR(){
		dir = NULL;
		pos = -2;	// Trick to fake .. and .
	}
};

static int litefsd_translate (
	LITEFSD_CTX &ctx,
	LITE_REQUEST_DATA &req,
	char path[PATH_MAX])
{
	int ret = -1;
	if (req.len > 0){
		strcpy (path,req.buf);
		ret = 0;
	}else{
		litefsd_sendfail (ctx);
	}
	return ret;
}

static const char *ondemandcmd = NULL;

static int litefsd_walk (const char *abspath, DIRDESC *&dir, FILEDESC *&file)
{
	int ret = -1;
	dir = NULL;
	file = NULL;
	int len = strlen(abspath);
	char tmp[len+1];
	strcpy (tmp,abspath);
	{
		int last = len - 1;
		if (last > 0 && strcmp(&tmp[last-1],"/.")==0){
			tmp[last-1] = '\0';
		}else if (last >= 0 && tmp[last] == '/'){
			tmp[last] = '\0';
		}
	}
	hosts.loadif("");
	if (tmp[0] == '\0'){
		dir = &hosts;
		ret = 0;
	}else{
		dir = &hosts;
		char *pt = tmp + 1;
		while (*pt != '\0'){
			char *start = pt;
			pt = strchr(pt,'/');
			n_debug ("stat recur start=:%s: pt=%p\n",start,pt);
			if (pt != NULL){
				*pt = '\0';
				DIRDESC *sdir=dir->lookupdir(start);
				if (sdir!=NULL){
					dir = sdir;
					dir->loadif(tmp+1);
				}else{
					break;
				}
				*pt++ = '/';
			}else{
				// We do two passes in case the ondemand command
				// is used
				for (int i=0; i<2; i++){
					DIRDESC *sdir=dir->lookupdir(start);
					if (sdir != NULL){
						dir = sdir;
						dir->loadif(tmp+1);
						ret = 0;
						break;
					}else{
						file = dir->lookupfile(start);
						if (file != NULL){
							ret = 0;
							break;
						}else if (ondemandcmd != NULL && i==0){
							char cmd[PATH_MAX];
							snprintf (cmd,sizeof(cmd)-1,"%s %s",ondemandcmd,tmp+1);
							system (cmd);
							char tmpcar = *start;
							*start = '\0';
							dir->load(tmp+1);
							*start = tmpcar;
						}
					}
				}
				break;
			}
		}
	}
	return ret;
}


/*
	Read a request from one client
*/
static int litefsd_doreq (
	BUFINPUT_REQUEST &buf_in,
	int fout,
	bool &redoset,		// Redo the set for select because ctls
						// have changes
	bool &shutdown)
{
	const int maxdir = 1024;
	static BROWSE_DIR *tbdir[maxdir];
	LITE_REQUEST_DATA req;
	int ret = buf_in.fill() > 0 ? 0 : -1;
	int len;
	while ((len = buf_in.read(req))>0){
		LITEFSD_CTX ctx;
		ctx.client_handle = req.client_handle;
		if (ctx.client_handle == -1) ctx.client_handle = fout;
		ctx.fd = fout;
		n_debug ("request %d\n",req.cmd);
		if (req.cmd == LITE_VERSION){
			if (req.mode != LITEFS_PROTOCOL_VERSION){
				litefsd_sendfail (ctx);
			}else{
				litefsd_sendok(ctx);
			}
		}else if (req.cmd == LITE_SHUTDOWN){
			litefsd_sendok(ctx);
			shutdown = true;
		}else if (req.cmd == LITE_STAT || req.cmd == LITE_LSTAT){
			n_debug ("stat :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,abspath)!=-1){
				n_debug ("stat abspath :%s:\n",abspath);
				DIRDESC *dir;
				FILEDESC *file;
				if (litefsd_walk (abspath,dir,file)!=-1){
					if (file != NULL){
						browse_sendstat (true,ctx,file->uid,file->gid,file->size);
					}else if (dir != NULL){
						browse_sendstat (false,ctx,getuid(),getgid(),1000);
					}else{
						litefsd_sendfail (ctx,ENOENT);
					}
				}else{
					litefsd_sendfail (ctx,ENOENT);
				}
			}
		}else if (req.cmd == LITE_CHDIR){
			// chdir means, may chdir
			char abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,abspath)!=-1){
				n_debug ("chdir abspath :%s:\n",abspath);
				DIRDESC *dir;
				FILEDESC *file;
				if (litefsd_walk (abspath,dir,file)!=-1
					&& file == NULL){
					litefsd_sendok (ctx);
				}else{
					litefsd_sendfail (ctx,ENOENT);
				}
			}
		}else if (req.cmd == LITE_UNLINK){
			char abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,abspath)!=-1){
				n_debug ("unlink abspath :%s:\n",abspath);
				DIRDESC *dir;
				FILEDESC *file;
				if (litefsd_walk (abspath,dir,file)!=-1
					&& file != NULL){
					char cmd[PATH_MAX];
					snprintf (cmd,sizeof(cmd)-1,"%s --unlink %s"
						,browsecmd,abspath+1);
					n_debug ("Executing :%s:\n",cmd);
					dir->last_loaded = 0;
					if (system(cmd)==0){
						litefsd_sendok (ctx);
					}else{
						litefsd_sendfail (ctx,EPERM);
					}
				}else{
					litefsd_sendfail (ctx,ENOENT);
				}
			}
		}else if (req.cmd == LITE_ACCESS){
			char abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,abspath)!=-1){
				n_debug ("access abspath :%s:\n",abspath);
				DIRDESC *dir;
				FILEDESC *file;
				if (litefsd_walk (abspath,dir,file)!=-1){
					// Incomplete. We must check ownership of the file
					// Just a test
					litefsd_sendok (ctx);
				}else{
					litefsd_sendfail (ctx,ENOENT);
				}
			}
		}else if (req.cmd == LITE_OPENDIR){
			char abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,abspath)!=-1){
				n_debug ("opendir abspath :%s:\n",abspath);
				int i;
				BROWSE_DIR *dirctrl = NULL;
				for (i=0; i<maxdir; i++){
					if (tbdir[i] == NULL){
						dirctrl = new BROWSE_DIR;
						tbdir[i] = dirctrl;
						break;
					}
				}
				DIRDESC *dir;
				FILEDESC *file;
				if (i==maxdir){
					litefsd_sendfail (ctx);
				}else if (litefsd_walk (abspath,dir,file)!=-1
					&& dir != NULL){
					dirctrl->dir = dir;
					litefsd_sendhandle (ctx,i);
				}else{
					delete dirctrl;
					tbdir[i] = NULL;
					litefsd_sendfail (ctx,ENOENT);
				}
			}
		}else if (req.cmd == LITE_READDIR){
			n_debug ("readir handle %d\n",req.handle);
			BROWSE_DIR *dirctrl = tbdir[req.handle];
			if (dirctrl == NULL){
				litefsd_sendfail (ctx);
			}else{
				LITE_REPLY_DATA rep;
				rep.status = LITE_STATUS_OK;
				rep.len = 0;
				DIRDESC *dir = dirctrl->dir;
				if (dir != NULL){
					NAMEOBJ *obj = NULL;
					if (dirctrl->pos == -2){
						// fake ..
						static NAMEOBJ ptpt("..");
						obj = &ptpt;
					}else if (dirctrl->pos == -1){
						// fake .
						static NAMEOBJ pt(".");
						obj = &pt;
					}else if (dirctrl->pos < dir->subdirs.nb){
						obj = dir->subdirs.tb[dirctrl->pos];
					}else if (dirctrl->pos < dir->subdirs.nb + dir->files.nb){
						obj = dir->files.tb[dirctrl->pos - dir->subdirs.nb];
					}
					if (obj != NULL){
						const char *name = obj->name;
						n_debug ("readdir name=:%s:\n",name);
						struct dirent ent;
						ent.d_ino = dirctrl->pos;
						ent.d_off = dirctrl->pos;
						ent.d_reclen = strlen(name);
						strcpy (ent.d_name,name);
						rep.len = sizeof(struct dirent);
						memcpy (rep.buf,&ent,sizeof(struct dirent));
						dirctrl->pos++;
					}
				}
				rep.client_handle = ctx.client_handle;
				write (fout,&rep,sizeof(LITE_REPLY)+rep.len);
			}
		}else if (req.cmd == LITE_CLOSEDIR){
			n_debug ("closedir %d\n",req.handle);
			BROWSE_DIR *dir = tbdir[req.handle];
			if (dir == NULL){
				litefsd_sendfail (ctx);
			}else{
				delete dir;
				tbdir[req.handle] = NULL;
				litefsd_sendok(ctx);
			}
		}else if (req.cmd == LITE_REWINDDIR){
			n_debug ("rewinddir %d\n",req.handle);
			BROWSE_DIR *dir = tbdir[req.handle];
			if (dir == NULL){
				litefsd_sendfail (ctx);
			}else{
				dir->pos = -2;
				litefsd_sendok(ctx);
			}
		}else if (req.cmd == LITE_OPEN){
			char abspath[PATH_MAX];
			n_debug ("browse Open\n");
			if (litefsd_translate (ctx,req,abspath)!=-1){
				n_debug ("browse Open :%s:\n",abspath);
			}
			errno = ENOENT;
			litefsd_senderrno(ctx);
		}else{
			fprintf (stderr,"litefsd-browse: Command %d not managed\n",req.cmd);
			litefsd_sendfail(ctx);
			ret = -1;
		}
	}
	return ret;
}
		
/*
	Process all request and connections. Infinite loop
*/
static void litefsd_loop (
	TBSOCK &tbsock)		// Unix/Tcp socket to receive new connection
{
	BUFINPUT_REQUEST *tbfd[1024];
	int nbfd = 0;
	fd_set fdin;
	FD_ZERO (&fdin);
	int maxfd = tbsock.fdset (fdin,0);
	bool shutdown = false;
	while (!shutdown){
		fd_set in;
		memcpy (&in,&fdin,sizeof(in));
		struct timeval timeout;
		timeout.tv_sec = 2000;
		timeout.tv_usec = 0;
		if (select (maxfd,&in,NULL,NULL,&timeout)!=-1){
			int orig_nbfd = nbfd;
			bool redo_fdin = false;
			for (int i=0; i<tbsock.nbsock; i++){
				int sock = tbsock.tb[i].handle;
				if (FD_ISSET (sock,&in)){
					struct sockaddr_un unc;
					size_t len = sizeof(unc);
					unc.sun_family = tbsock.tb[i].family;
					int s = accept (sock,(struct sockaddr*)&unc,&len);
					tbfd[nbfd++] = new BUFINPUT_REQUEST(s);
					FD_SET (s,&fdin);
					if (s >= maxfd) maxfd = s + 1;
				}
			}
			for (int i=0; i< orig_nbfd; i++){
				BUFINPUT_REQUEST *ptfd = tbfd[i];
				int fd = ptfd->fd;
				if (FD_ISSET(fd,&in)){
					if (litefsd_doreq (*ptfd,fd,redo_fdin,shutdown) == -1){
						close (fd);
						redo_fdin = true;
						delete ptfd;
						tbfd[i] = NULL;
					}
				}
			}
			if (redo_fdin){
				FD_ZERO (&fdin);
				maxfd = tbsock.fdset (fdin,0);
				int copyfd = 0;
				for (int i=0; i<nbfd; i++){
					if (tbfd[i] != NULL){
						int fd = tbfd[i]->fd;
						FD_SET (fd,&fdin);
						if (fd >= maxfd) maxfd = fd + 1;
						tbfd[copyfd++] = tbfd[i];
					}
				}
				nbfd = copyfd;
			}
		}
	}
}

int main (int argc, char *argv[])
{
	int ret = -1;
	if (argc == 1){
		usage();
	}else{
		TBSOCK tbsock;		// Listening sockets
		bool dofork = false;
		const char *pidfile = NULL;
		const char *sockenv = getenv ("LITEFS_SOCKET");
		if (sockenv != NULL){
			int sock = litefsd_server(sockenv);
			if (sock != -1){
				tbsock.add (AF_UNIX,sock);
			}
		}
		for (int i=1; i<argc; i++){
			if (strcmp(argv[i],"--socket")==0){
				i++;
				int sock = litefsd_server(argv[i]);
				if (sock != -1){
					tbsock.add (AF_UNIX,sock);
				}
			}else if (strcmp(argv[i],"--debug")==0){
				debug = true;
			}else if (strcmp(argv[i],"--fork")==0){
				dofork = true;
			}else if (strcmp(argv[i],"--pidfile")==0){
				i++;
				pidfile = argv[i];
			}else if (strcmp(argv[i],"--browsecmd")==0){
				i++;
				browsecmd = argv[i];
			}else if (strcmp(argv[i],"--ondemandcmd")==0){
				i++;
				ondemandcmd = argv[i];
			}else{
				usage();
				exit (-1);
			}
		}
		char cwd[PATH_MAX];
		if (getcwd(cwd,PATH_MAX)==NULL){
			fprintf (stderr,"Can't locate current working directory (%s)\n"
				,strerror(errno));
		}else{
			if (tbsock.nbsock > 0){
				if (dofork){
					if (fork() != 0) exit(0);
					setsid();
				}
				if (pidfile != NULL){
					FILE *fout = fopen (pidfile,"w");
					if (fout != NULL){
						fprintf (fout,"%d\n",getpid());
						fclose (fout);
					}
				}
				litefsd_loop(tbsock);
			}else{
				usage();
			}
		}
	}
	return ret;
}




