#include <stdio.h>
#include <errno.h>
#include <stdarg.h>
#include <unistd.h>
#include <limits.h>
#include <ctype.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/wait.h>
#include <netdb.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include "../tcpaccess.h"

static void usage()
{
	fprintf (stderr,"rprintserv version 1.0\n"
		"\n"
		"Common options: --prtdev /dev/file\n"
		"                --debug\n"
		"                --pidfile path\n"
		"                --nodaemon\n"
		"\n"
		"                --tcpallow ip\n"
		"                --tcpallow net/mask\n"
		"                --secureport\n"
		);
}

static bool debug=false;

static void n_debug (const char *ctl, ...)
{
	if (debug){
		fprintf (stderr,"|%d| ",getpid());
		va_list list;
		va_start (list,ctl);
		vfprintf (stderr,ctl,list);
		va_end (list);
	}
}


static pid_t tbend[10];
static int nbend;

static void fctchild (int )
{
	int status;
	pid_t pid = wait(&status);
	n_debug("fctchild %d %d\n",pid,status);
	tbend[nbend] = pid;
	nbend = (nbend+1)%10;
}

/*
	Open the printer port (to listen for connection)
	Return the socket or -1 if any error
*/
int rprint_openport()
{
	int ret = -1;
	int port = -1;
	static const char *portstr = "printer";
	struct servent *serv = getservbyname (portstr,"tcp");
	if (serv != NULL){
		port = serv->s_port;
	}
	if (port == -1){
		fprintf (stderr,"tcp port unknown %s\n",portstr);
	}else{
		struct sockaddr_in sin;
		sin.sin_family = AF_INET;
		memset  (&sin.sin_addr,0,sizeof(sin.sin_addr));
		sin.sin_port = port;

		for (int i=0; i<5; i++){
			int listen_handle = socket (AF_INET, SOCK_STREAM, 0);
			sin.sin_addr.s_addr = INADDR_ANY;
			int opt = 1;
			if (listen_handle == -1){
				fprintf (stderr,"listen_handle %d(%s)\n"
					,errno
					,strerror(errno));
			}else if (setsockopt(listen_handle,SOL_SOCKET,SO_REUSEADDR,&opt,sizeof(opt))==-1){
				fprintf (stderr,"Can't set socket option SO_REUSEADDR (%s)\n"
					,strerror (errno));
			}else if (bind (listen_handle,(struct sockaddr *) &sin, sizeof (sin)) == -1){
				fprintf (stderr,"bind %d(%s)\n",errno
					,strerror(errno));
			}else if (listen (listen_handle,5) == -1){
				fprintf (stderr,"listen %d(%s)\n",errno
					,strerror(errno));
				break;
			}else{
				ret = listen_handle;
				break;
			}
			close (listen_handle);
			listen_handle = -1;
			if (i < 5) sleep (i*5);
		}
	}
	return ret;
}

static int rprint_openprt (
	const char *dev_prefix,		// /dev/lp
	int nbdev,					// try /dev/lp0 /dev/lp1 ... up to nbdev
	int tbprt[],				// Put the opened file handle here
	int nbprt,					// Current handle in tbprt[]
	int sizetb)					// Size of table
{
	for (int i=0; i<nbdev; i++){
		char path[PATH_MAX];
		sprintf (path,"%s%d",dev_prefix,i);
		int fd = open (path,O_WRONLY);
		if (fd != -1){
			printf ("Serving queue lp%d on device %s\n",nbprt,path);
			tbprt[nbprt++] = fd;
			if (nbprt == sizetb) break;
		}
	}
	return nbprt;
}
/*
	Probe an open printer ports
	Return the number of handle placed in tbprt[]
*/
static int rprint_openprt (int tbprt[], int sizetb)
{
	int nbprt = rprint_openprt ("/dev/lp",3,tbprt,0,sizetb);
	return rprint_openprt ("/dev/usb/lp",16,tbprt,nbprt,sizetb);
}

struct HANDLE_INFO {
	int fd;
	int noprt;	// On which printer we must print
	pid_t pid;
	HANDLE_INFO(int _fd){
		reset();
		fd = _fd;
	}
	void reset(){
		fd = -1;
		noprt = -1;
		pid = 0;
	}
};


static int rprint_setfd (fd_set &in, int sock, HANDLE_INFO *tb[], int nbhandle)
{
	FD_ZERO (&in);
	int maxfd = sock;
	FD_SET (sock,&in);
	for (int i=0; i<nbhandle; i++){
		HANDLE_INFO *info = tb[i];
		if (info->noprt == -1){
			int fd = info->fd;
			if (fd != -1){
				FD_SET (fd,&in);
				if (fd > maxfd) maxfd = fd;
			}
		}
	}
	return maxfd+1;
}
/*
	Send a positive acknowledge
*/
static void rprint_ack (int fd)
{
	static char ch=0;
	write (fd,&ch,1);
}

static void rprint_dump (const char *s, const char *buf, int len)
{
	if (debug){
		fprintf (stderr,"|%d| %s[%d]:",getpid(),s,len);
		for (int j=0; j<len; j++) fprintf (stderr," %02x",buf[j]);
		fprintf (stderr,"\n");
	}
}

static int nprint_readexit (int fd, char *buf, int size)
{
	int ret = read (fd,buf,size);
	if (ret <= 0) _exit(0);
	return ret;
}

static int rprint_do (int fd, int prtfd)
{
	pid_t ret = fork();
	/*
		We receive the data to print
		The line goes like this
		\2PRT\n
		\3NNN FILE\n
		data
		\0
		\2NNN control-file
		data
		\0

		NNN is the size of the data
		FILE is the name of the df file
		(we do not care)
		PRT name of the print queue
	*/
	if (ret == 0){
		while (1){
			// Read the data and discard the control file
			char buf[1000];
			int len = nprint_readexit (fd,buf,999);
			rprint_dump ("datafile",buf,len);
			buf[len] = '\0';
			if (buf[0] != 3 && buf[0] != 2) _exit(0);
			bool is_data = buf[0] == 3;
			int size = atoi(buf+1);
			char *pt = buf+1;
			while (isdigit(*pt)) pt++;
			if (*pt != ' ') _exit (0);
			#if 0
				pt++;
				// We skip the name of the df file as we do not need it.
				// We print directly to the port
				while (*pt != '\n' && *pt != '\0') pt++;
			#endif
			rprint_ack(fd);
			n_debug ("Expect %d bytes\n",size);
			while (size > 0){
				int len = nprint_readexit (fd,buf,size >= 1000 ? 1000 : size);
				if (len <= 0) break;
				n_debug ("read data %d\n",len);
				if (is_data) write (prtfd,buf,len);
				#if 0
					for (int i=0; i<len; i++){
						write (prtfd,buf+i,1);
						sleep(1);
					}
				#endif
				size -= len;
			}
			n_debug ("Wait for data ack\n");
			len = nprint_readexit (fd,buf,1);
			rprint_dump ("datafile2",buf,len);
			n_debug ("Sending data ack\n");
			rprint_ack(fd);
		}
		close (fd);
		_exit (0);
	}
	return ret;
}

static int rprint_queue2dev (const char *buf)
{
	char queue[strlen(buf)];
	char *dst = queue;
	while (*buf != '\0' && *buf != '\n') *dst++ = *buf++;
	*dst = '\0';
	n_debug ("Sending to queue :%s:\n",queue);
	int ret = -1;
	if (strcasecmp(queue,"lp")==0){
		ret = 0;
	}else if (strncasecmp(queue,"lp",2)==0){
		ret = atoi(queue+2);
	}
	return ret;
}

static void rprint_dispatch (
	int tbprt[],		// Handle to printers
	int tbpid[],		// Pid of the currently printing jobs
						// for each printers or 0
	HANDLE_INFO *tb[],	// Potential requests
	int nb)
{
	for (int i=0; i<nb; i++){
		HANDLE_INFO *info = tb[i];
		int noprt = info->noprt;
		if (noprt != -1){
			if (tbpid[noprt] == 0){
				info->pid = rprint_do (info->fd,tbprt[noprt]);
				n_debug ("Dispatching handle %d to pid %d\n",info->fd,info->pid);
				tbpid[noprt] = info->pid;
				#if 1
					n_debug ("dispatch closing %d\n",info->fd);
					close (info->fd);
					info->fd = -1;
				#endif
			}
		}
	}
}

static void rprint_sendstat (
	int fd,
	HANDLE_INFO *tb[],
	int nbhandle,
	pid_t tbpid[],
	int noprt)
{
	bool one = false;
	if (tbpid[noprt] != 0){
		char line[200];
		int len = sprintf (line,"process %d active\n",tbpid[noprt]);
		write (fd,line,len);
		one = true;
	}
	for (int i=0; i<nbhandle; i++){
		HANDLE_INFO *info = tb[i];
		if (info->noprt == noprt){
			char line[200];
			int len = sprintf (line,"job %d\n",i);
			write (fd,line,len);
			one = true;
		}
	}
	if (!one){
		write (fd,"no entries\n",11);
	}
}

/*
	Wait for print request and status request
	
*/
static void rprint_loop (
	int sock,
	int tbprt[],
	int nbprt,
	TCPACCESS &tcpa)
{
	// Forked process handling one print request
	pid_t tbpid[nbprt];
	memset (tbpid,0,sizeof(tbpid));
	HANDLE_INFO *tbhandle[100];
	int nbhandle = 0;
	signal (SIGCHLD,fctchild);
	while (1){
		fd_set fdin;
		int maxfd = rprint_setfd (fdin,sock,tbhandle,nbhandle);
		int sel = select (maxfd,&fdin,NULL,NULL,NULL);
		n_debug ("select %d\n",sel);
		// Some childs may have ended
		for (int i=0; i<nbprt; i++){
			if (tbpid[i] != 0){
				pid_t pid = tbpid[i];
				for (int j=0; j<10; j++){
					if (tbend[j] == pid){
						tbpid[i] = 0;
						for (int h=0; h<nbhandle; h++){
							HANDLE_INFO *info = tbhandle[h];
							if (info->pid == pid){
								info->pid = 0;
								info->noprt = -1;
								break;
							}
						}
						// Trying to dispatch another print job
						n_debug ("Trying to dispatch %d\n",i);
						rprint_dispatch (tbprt,tbpid,tbhandle,nbhandle);
						break;
					}
				}
			}
		}
		if (sel > 0){
			if (FD_ISSET(sock,&fdin)){
				// New connection
				struct sockaddr_un unc;
				size_t len = sizeof(unc);
				unc.sun_family = AF_INET;
				int fd = accept (sock,(struct sockaddr*)&unc,&len);
				if (fd != -1){
					if (!tcpa.isok(fd)){
						close (fd);
					}else{
						tbhandle[nbhandle++] = new HANDLE_INFO(fd);
					}
				}
			}
			for (int i=0; i<nbhandle; i++){
				HANDLE_INFO *info = tbhandle[i];
				int fd = info->fd;
				if (info->noprt == -1 && FD_ISSET(fd,&fdin)){
					char buf[1000];
					int len = read (fd,buf,999);
					bool must_close = false;
					if (len <= 0){
						must_close = true;
					}else{
						buf[len] = '\0';
						if (buf[0] == 0){
							// Don't know
							rprint_dump ("don't know 0",buf,len);
							rprint_ack (fd);
							must_close = true;
						}else if (buf[0] == 1){
							// Resume printing. Do nothing for now
							rprint_dump ("resume",buf,len);
							// must_close = true;
						}else if (buf[0] == 2){
							// Print a file
							rprint_dump ("printfile",buf,len);
							int noprt = rprint_queue2dev (buf+1);
							rprint_ack (fd);
							if (noprt >=0 && noprt < nbprt){
								info->noprt = noprt;
								rprint_dispatch (tbprt,tbpid,tbhandle,nbhandle);
							}else{
								fprintf (stderr,"Invalid print queue %d\n",noprt);
								must_close = true;
							}
						}else if (buf[0] == 3){
							// Send a queue report. short
							rprint_dump ("queue report short",buf,len);
							write (fd,"no entries\n",11);
							must_close = true;
						}else if (buf[0] == 4){
							// Send a queue report. long
							rprint_dump ("queue report long",buf,len);
							int noprt = rprint_queue2dev (buf+1);
							if (noprt >=0 && noprt < nbprt){
								rprint_sendstat (fd,tbhandle,nbhandle,tbpid,noprt);
							}else{
								write (fd,"no entries\n",11);
							}
							must_close = true;
						}else if (buf[0] == 5){
							// Remove a job from the queue
							rprint_dump ("remove a job",buf,len);
							rprint_ack(fd);
							must_close = true;
						}else{
							fprintf (stderr,"Invalid request %02x\n",buf[0]);
							must_close = true;
						}
					}
					if (must_close){
						n_debug ("Closing handle %d\n",fd);
						close (fd);
						info->fd = -1;
					}
				}
			}
		}
		int copyto = 0;
		for (int i=0; i<nbhandle; i++){
			HANDLE_INFO *info = tbhandle[i];
			if (info->fd != -1){
				tbhandle[copyto++] = info;
			}else{
				delete info;
			}
		}
		nbhandle = copyto;
	}
}


int main (int argc, char *argv[])
{
	int ret = -1;
	TCPACCESS tcpa;		// Which hosts or network may connect
	const int MAX_PRINTERS=50;
	int tbprt[MAX_PRINTERS];	// File handle to print devices
	int nbprt = 0;
	bool nodaemon = false;
	for (int i=1; i<argc; i++){
		if (strcmp(argv[i],"--tcpallow")==0){
			i++;
			tcpa.add (argv[i]);
		}else if (strcmp(argv[i],"--prtdev")==0){
			i++;
			if (nbprt == MAX_PRINTERS){
				fprintf (stderr,"Too many printers, maximum %d\n",MAX_PRINTERS);
			}else{
				int fd = open (argv[i],O_WRONLY);
				if (fd == -1){
					fprintf (stderr,"Can't open printer device %s (%s)\n"
						,argv[i],strerror(errno));
				}else{
					tbprt[nbprt++] = fd;
				}
			}
		}else if (strcmp(argv[i],"--pidfile")==0){
			i++;
			FILE *fout = fopen (argv[i],"w");
			if (fout != NULL){
				fprintf (fout,"%d\n",getpid());
				fclose (fout);
			}
		}else if (strcmp(argv[i],"--debug")==0){
			debug = true;
		}else if (strcmp(argv[i],"--secureport")==0){
			tcpa.requestsecureport();
		}else if (strcmp(argv[i],"--nodaemon")==0){
			nodaemon = true;
		}else{
			usage();
			exit (-1);
		}
	}
	if (nbprt == 0) nbprt = rprint_openprt (tbprt,MAX_PRINTERS);
	if (nbprt <= 0){
		fprintf (stderr,"No printer defined\n");
	}else{
		int sock = rprint_openport();
		if (sock != -1){
			if (nodaemon || fork()==0){
				rprint_loop (sock,tbprt,nbprt,tcpa);
			}
			ret = 0;
		}
	}
	return ret;
}



